/*************************************************************************
 *    C H I P - 8     * CHIP8 is (C) Copyright Matt L. 2021              *
 *     (C) Matt       * WWW: <https://www.helloboi.tk/chip8>             *
 * <www.helloboi.tk>  * EMAIL: <diskoperatingsystem@protonmail.com>      *
 *************************************************************************
 * This file is part of CHIP8.                                           *
 *                                                                       *
 * CHIP8 is free software: you can redistribute it and/or modify         *
 * it under the terms of the GNU General Public License as published by  *
 * the Free Software Foundation, either version 3 of the License, or     *
 * (at your option) any later version.                                   *
 *                                                                       *
 * CHIP8 is distributed in the hope that it will be useful,              *
 * but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 * GNU General Public License for more details.                          *
 *                                                                       *
 * You should have received a copy of the GNU General Public License     *
 * along with CHIP8.  If not, see <https://www.gnu.org/licenses/>.       *
 *************************************************************************/
#include <SDL2/SDL.h>
#include <stdio.h>
#include <stdint.h>
#include "config.h"

/* screen[y][x] */
uint8_t screen[32][64];
/* V0-VF */
uint8_t registers[16];
/* I(ndex) register */
uint16_t register_i;
/* program counter */
uint16_t pc;
/* memory (programs start at 0x200) */
uint8_t memory[4096];
/* return stack */
uint16_t stack[64];
/* stack pointer */
uint8_t sp = -1;
/* timers */
uint8_t delay, sound;
/* SDL variables */
SDL_Window *window = NULL;
SDL_Surface *surface = NULL;
/* time of previous tick */
Uint32 prev = 0;
/* representing key */
const char *keystr;

/*****************************************
 * leave_sdl:                            *
 * close SDL windows and stop SDL system *
 *****************************************/
void leave_sdl(void)
{
    SDL_DestroyWindow(window);
    SDL_Quit();
}

/*******************************
 * dump_registers:             *
 * print registers to stdout   *
 *******************************/
void dump_registers(void)
{
    int i;
    fprintf(stderr, "pc=%x i=%x ", (unsigned int)pc, (unsigned int)register_i);
    for (i=0;i<16;i++)
        fprintf(stderr, "v%x=%x ", i, (unsigned int)registers[i]);
    fputc('\n', stderr);
}


/********************************************************
 * error:                                               *
 * close windows, dump registers, print s and terminate *
 ********************************************************/
void error(const char *s)
{
    SDL_Delay(1000);
    leave_sdl();
    dump_registers();
    fputs(s, stderr);
    fputc('\n', stderr);
    exit(EXIT_FAILURE);
}

/*****************
 * clear_screen: *
 * clear screen  *
 *****************/
void clear_screen(void)
{
    int i, j;
    for (i=0;i<32;i++)
        for (j=0;j<64;j++)
            screen[i][j] = 0;
}

/******************************************************
 * write_screen:                                      *
 * refresh the screen with the screen[y][x] structure *
 ******************************************************/
void write_screen(void)
{
    SDL_Rect pixel;
    int i, j;
    pixel.x = 0;
    pixel.y = 0;
    /* window = 640x480
     * CHIP-8 screen = 64x32
     * 640/64 = 10 = width
     * 480/32 = 15 = height
     */
    pixel.w = 10;
    pixel.h = 15;
    SDL_UnlockSurface(surface);
    for (i=0;i<32;i++)
        for (j=0;j<64;j++)
        {
            pixel.x = j*10;
            pixel.y = i*15;
            if (screen[i][j])
                SDL_FillRect(surface, &pixel, SDL_MapRGB(surface->format, 0xFF, 0xFF, 0xFF));
            else
                SDL_FillRect(surface, &pixel, SDL_MapRGB(surface->format, 0x00, 0x00, 0x00));
        }
    if (SDL_UpdateWindowSurface(window)) error(SDL_GetError());
    SDL_LockSurface(surface);
}

/************************************************
 * draw_sprite:                                 *
 * draw a sprite at screen[y][x] with size size *
 * return 1 if pixels unset else 0              *
 ************************************************/
int draw_sprite(int y, int x, int size)
{
    /* i, j = iterators
     * k = counter
     * l = return value
     */
    int i, j, k, l = 0;
    for (i=0;i<size;i++)
    {
        k = 0;
        for (j=0x80;j;j=j>>1)
        {
            /* set flag */
            if ((memory[register_i + i]& j) && !(screen[(y + i) % 32][(x + k) % 64] ^ ((memory[register_i + i] & j) != 0)))
                l = 1;
            screen[(y + i) % 32][(x + k) % 64] ^= (memory[register_i + i] & j) != 0;
            k++;
        }
    }
    write_screen();
    return l;
}

/***************************
 * reset:                  *
 * reset the CHIP-8 system *
 ***************************/
void reset(void)
{
    int i;
    clear_screen();
    for (i=0;i<16;i++)
        registers[i] = 0;
    register_i = 0;
    pc = 0x200;
    sp = -1;
    write_screen();
    keystr = NULL;
    /* copy fonts to memory */
    for (i=0;i<16;i++)
        memcpy(&memory[i*5], &font[i][0], sizeof(uint8_t)*5);
    prev = SDL_GetTicks();
}

/**********************************
 * init:                          *
 * start SDL, reset CHIP-8 system *
 **********************************/
void init(void)
{
    /* SDL functions return a negative value on failure */
    if (SDL_Init(SDL_INIT_VIDEO) < 0)
        error(SDL_GetError());
    if ((window = SDL_CreateWindow("CHIP-8 Interpreter", SDL_WINDOWPOS_UNDEFINED,
                                  SDL_WINDOWPOS_UNDEFINED, 640, 480, SDL_WINDOW_SHOWN)) == NULL)
        error(SDL_GetError());
    surface = SDL_GetWindowSurface(window);
    SDL_UnlockSurface(surface);
    reset();
}

/****************************************************
 * getch:                                           *
 * non-blocking check for keypress                  *
 * returns char if keypress available else return 0 *
 ****************************************************/
int getch(void)
{
    if (keystr == NULL)
        return 0;
    /* non alphanumeric character, just return 0 */
    if (keystr[1])
    {
        keystr = NULL;
        return 0;
    }
    else
    {
        char s = *keystr;
        keystr = NULL;
        return tolower(s);
    }
    keystr = NULL;
    return 0;
}

/************************************************
 * key_to_map:                                  *
 * translate "real" keypress to mapped keypress *
 ************************************************/
char key_to_map(char c)
{
    int i;
    for (i=0;i<16;i++)
        if (keymap[i] == c)
            return keymap_to[i];
    return -1;
}

/**********************
 * opcode:            *
 * execute one opcode *
 **********************/
void opcode(void)
{
    int i, op = memory[pc++];
    switch ((op & 0xF0) >> 4)
    {
        /* execute machine code routine (unimplemented except for 2 routines) */
        case 0x00:
            if ((op & 0x0F) == 0)
                switch (memory[pc++])
                {
                    case 0xE0:
                        clear_screen();
                        write_screen();
                        break;
                    case 0xEE:
                        pc = stack[sp--];
                        break;
                    default:
                        error("machine code execution not implemented");
                        break;
                }
            else
                error("machine code execution not implemented");
            break;
        /* jump to address */
        case 0x01:
            pc = ((op & 0x0F) << 4) << 4 | memory[pc];
            break;
        /* branch to subroutine */
        case 0x02:
            stack[++sp] = pc + 1;
            pc = ((op & 0x0F) << 4) << 4 | memory[pc];
            break;
        /* skip next instruction if VX equal to NN
         * 3XNN
         */
        case 0x03:
            if (registers[op & 0x0F] == memory[pc++])
                pc += 2;
            break;
        /* skip next instruction if VX NOT equal to NN
         * 4XNN
         */
        case 0x04:
            if (registers[op & 0x0F] != memory[pc++])
                pc += 2;
            break;
        /* skip next instruction if VX equal to VY
         * 5XY0
         */
        case 0x05:
            if (registers[op & 0x0F] == registers[memory[pc++] >> 4])
                pc += 2;
            break;
        /* set VX to NN */
        case 0x06:
            registers[op & 0x0F] = memory[pc++];
            break;
        /* add NN to VX */
        case 0x07:
            registers[op & 0x0F] += memory[pc++];
            break;
        case 0x08:
            switch (memory[pc] & 0x0F)
            {
                /* store VY in VX */
                case 0:
                    registers[op & 0x0F] = registers[(memory[pc++] & 0xF0) >> 4];
                    break;
                /* VX |= VY */
                case 1:
                    registers[op & 0x0F] |= registers[(memory[pc++] & 0xF0) >> 4];
                    break;
                /* VX &= VY */
                case 2:
                    registers[op & 0x0F] &= registers[(memory[pc++] & 0xF0) >> 4];
                    break;
                /* VX ^= VY */
                case 3:
                    registers[op & 0x0F] ^= registers[(memory[pc++] & 0xF0) >> 4];
                    break;
                /* VX += VY */
                case 4:
                    i = registers[op & 0x0F];
                    registers[op & 0x0F] += registers[(memory[pc++] & 0xF0) >> 4];
                    /* set flag */
                    if (registers[op & 0x0F] < i)
                        registers[0x0F] = 1;
                    break;
                /* VX -= VY */
                case 5:
                    i = registers[op & 0x0F];
                    registers[op & 0x0F] -= registers[(memory[pc++] & 0xF0) >> 4];
                    /* set borrow flag */
                    if (i > registers[op & 0x0F])
                        registers[0x0F] = 1;
                    else
                        registers[0x0F] = 0;
                    break;
                /* VX = VY >> 1 */
                case 6:
                    /* set flag to LSB */
                    registers[0x0F] = registers[(memory[pc] & 0xF0) >> 4] & 1;
                    registers[op & 0x0F] = registers[(memory[pc++] & 0xF0) >> 4] >> 1;
                    break;
                /* VX = VY - VX */
                case 7:
                    i = registers[op & 0x0F];
                    registers[op & 0x0F] = registers[(memory[pc++] & 0xF0) >> 4] - i;
                    if (registers[op & 0x0F] < i)
                        registers[0x0F] = 1;
                    else
                        registers[0x0F] = 0;
                    break;
                /* VX = VY << 1 */
                case 8:
                    /* set flag to MSB */
                    registers[0x0F] = (registers[(memory[pc] & 0xF0) >> 4] & 0x80) != 0;
                    registers[op & 0x0F] = registers[(memory[pc++] & 0xF0) >> 4] << 1;
                    break;
                default:
                    error("unrecognized instruction");
                    break;
            }
            break;
        /* skip next instruction if VX NOT equal VY
         * 9XY0
         */
        case 0x09:
            if (registers[op & 0x0F] != registers[(memory[pc++] & 0xF0) >> 4])
                pc += 2;
            break;
        /* store NNN in I */
        case 0x0A:
            register_i = ((op & 0x0F) << 4 << 4) | memory[pc++];
            break;
        /* branch to NNN + V0 */
        case 0x0B:
            pc = ((op & 0x0F) << 4 << 4) | memory[pc] + registers[0];
            break;
        /* set VX to rand()&NN
         * CXNN
         */
        case 0x0C:
            registers[op & 0x0F] = rand() & memory[pc++];
            break;
        /* draw sprite at screen[VY][VX] with
         * N bytes of sprite data
         * VF=1 if pixels go from set to unset
         * DXYN
         */
        case 0x0D:
            /* draw_sprite(int y, int x, int size) */
            registers[0x0F] = draw_sprite(registers[(memory[pc] & 0xF0) >> 4],
                                          registers[op & 0x0F], memory[pc] & 0x0F);
            pc++;
            break;
        /* non-blocking check for keypress */
        case 0x0E:
            /* skip next instruction if VX = key */
            if (memory[pc] == 0x9E)
                /* prevent dangling else */
                if ((i = key_to_map(getch())) != -1)
                {
                    if (registers[op & 0x0F] == i)
                        pc += 3;
                    else
                        pc++;
                } else
                    pc++;
            else if (memory[pc] == 0xA1)
                if ((i = key_to_map(getch())) != -1)
                {
                    if (registers[op & 0x0F] != i)
                        pc += 3;
                    else
                        pc++;
                } else
                    pc++;
            else
                error("unrecognized instruction");
            break;
        case 0x0F:
            switch (memory[pc++])
            {
                case 0x07:
                    registers[op & 0x0F] = delay;
                    break;
                case 0x0A:
                    if ((i = getch()) == -1)
                        pc -= 2;
                    else
                        registers[op & 0x0F] = key_to_map(i);
                    break;
                case 0x15:
                    delay = registers[op & 0x0F];
                    break;
                case 0x18:
                    sound = registers[op & 0x0F];
                    break;
                case 0x1E:
                    register_i += registers[op & 0x0F];
                    break;
                case 0x29:
                    register_i = registers[op & 0x0F]*5;
                    break;
                case 0x33:
                    i = registers[op & 0x0F];
                    memory[register_i] = i/100;
                    memory[register_i+1] = (i/10)%10;
                    memory[register_i+1] = i%10;
                    break;
                case 0x55:
                    for (i=0;i<=(op&0x0F);i++)
                        memory[register_i + i] = registers[i];
                    break;
                case 0x65:
                    for (i=0;i<=(op&0x0F);i++)
                        registers[i] = memory[register_i++];
                    break;
            }
            break;
        default:
            error("unrecognized instruction");
            break;
    }
}

/******************
 * chip8_tick():  *
 * perform a tick *
 ******************/
void chip8_tick(void)
{
    SDL_Event event;
    if (SDL_GetTicks() > prev + 16)
    {
        if (sound) sound--;
        if (delay) delay--;
        prev = SDL_GetTicks();
    }
    opcode();
    while (SDL_PollEvent(&event))
        if (event.type == SDL_QUIT)
        {
            leave_sdl();
            exit(EXIT_FAILURE);
        }
        else if (event.type == SDL_KEYDOWN)
            keystr = SDL_GetKeyName(SDL_GetKeyFromScancode(event.key.keysym.scancode));
    if (SDL_UpdateWindowSurface(window))
        error(SDL_GetError());
}

/***********************************
 * chip8_run:                      *
 * begin running the CHIP-8 system *
 ***********************************/
void chip8_run(void)
{
    reset();
    for (;;) chip8_tick();
}

/*********************
 * info:             *
 * print information *
 *********************/
void info(void)
{
    puts("CHIP-8 interpreter  Copyright (C) 2021 Matt <www.helloboi.tk>");
    puts("This program comes with ABSOLUTELY NO WARRANTY.");
    puts("This is free software, and you are welcome to distribute it under certain conditions.");
    puts("Consult the LICENSE file for more details.");
}

/***************************
 * main:                   *
 * begin the CHIP-8 system *
 ***************************/
int main(int argc, char **argv)
{
    int i;
    info();
    if (argc == 1)
    {
        fputs("usage: chip8 file\n", stderr);
        return 1;
    }
    for (i=1;i<argc;i++)
        if (!strcmp(argv[i], "-h"))
        {
            fputs("usage: chip8 file\n", stderr);
            return 1;
        }
        else
        {
            FILE *input = fopen(argv[i], "r");
            if (input == NULL)
                error("could not open file");
            fread(&memory[0x200], 1, 4096-0x200, input);
        }
    init();
    chip8_run();
    SDL_Delay(1000);
    leave_sdl();
    return 0;
}
